import difflib
import logging
from collections import OrderedDict

from .actions import Favorite, Follow
from .client import TwitterClient, TwitterClientRetweetedException
from .config import TwitterConfig
from .ignorelist import IgnoreList
from .notifier import NotificationService
from .post_queue import PostQueue
from .scheduler import PeriodicScheduler

# The logger object
logger = logging.getLogger(__name__)


class Yatcobot():

    def __init__(self, ignore_list_file):

        self.ignore_list = IgnoreList(ignore_list_file)
        self.post_queue = PostQueue()
        self.client = TwitterClient(TwitterConfig.get()['consumer_key'],
                                    TwitterConfig.get()['consumer_secret'],
                                    TwitterConfig.get()['access_token_key'],
                                    TwitterConfig.get()['access_token_secret'])
        self.scheduler = PeriodicScheduler()
        self.notification = NotificationService()
        self.actions = [Follow(self.client), Favorite(self.client)]
        self.last_mention = None

    def enter_contest(self):
        """ Gets one post from post_queue and retweets it"""

        logger.info("=== CHECKING RETWEET QUEUE ===")

        logger.info("Queue length: {}".format(len(self.post_queue)))

        if len(self.post_queue) > 0:

            post_id, post = self.post_queue.popitem(last=False)

            text = post['full_text'].replace('\n', '')
            text = (text[:75] + '..') if len(text) > 75 else text
            logger.info("Retweeting: {0} {1}".format(post['id'], text))
            logger.debug("Tweet score: {}".format(post['score']))

            if post['user']['id'] in self.ignore_list:
                logger.info("Blocked user's tweet skipped")
                return
            try:
                self.client.retweet(post['id'])
                self.ignore_list.append(post['id'])
            except TwitterClientRetweetedException:
                self.ignore_list.append(post['id'])
                logger.error("Alredy retweeted tweet with id {}".format(post['id']))
                return

            for action in self.actions:
                action.process(post)

    def clear_queue(self):
        """Removes the extraneous posts from the post_queue"""

        to_delete = len(self.post_queue) - TwitterConfig.get().search.max_queue

        if to_delete > 0:
            for i in range(to_delete):
                # Remove from the end where the posts has lower score
                self.post_queue.popitem()

            logger.info("===THE QUEUE HAS BEEN CLEARED=== Deleted {} posts".format(to_delete))

    def update_blocked_users(self):
        """Gets the blocked users and adds their ids in the ignore list"""

        for b in self.client.get_blocks():
            if not b in self.ignore_list:
                self.ignore_list.append(b)
                logger.info("Blocked user {0} added to ignore list".format(b))

    def scan_new_contests(self):
        """Searches the twitter for new contests and adds the to the post queue"""

        logger.info("=== SCANNING FOR NEW CONTESTS ===")

        for search_query in TwitterConfig.get().search.queries:

            if isinstance(search_query, str):
                results = self.client.search_tweets(search_query, 50)
            elif isinstance(search_query, OrderedDict):
                lang = search_query['lang']
                search_query = list(search_query.keys())[0]

                results = self.client.search_tweets(search_query, 50, language=lang)
            else:
                raise ValueError("Uknown type of query {}".format(str(search_query)))

            logger.info("Got {} new results for: {}".format(len(results), search_query))

            for post in results:
                self._insert_post_to_queue(post)

        # Filter queue based on configure filters
        self.post_queue.filter()
        # Sort the queue based on some features
        self.post_queue.sort()

    def check_new_mentions(self):
        """
        Check if someone mentioned the user and sends a notification
        Usefull because many winners are mentioned in tweets
        """

        # Check if notification is enabled
        if not self.notification.is_enabled():
            return

        # If its the first time its called get the last mention
        logger.info("=== CHECKING NEW MENTIONS ===")
        if self.last_mention is None:
            posts = self.client.get_mentions_timeline(count=1)
            if len(posts) > 0:
                self.last_mention = posts[0]
            return

        # Else check if there are new mentions after the last, notify
        posts = self.client.get_mentions_timeline(since_id=self.last_mention['id'])
        if len(posts) > 0:
            links = ' , '.join(self.create_tweet_link(x) for x in posts)
            logger.info("You ve got {} new mentions: {}".format(len(posts), links))
            self.notification.send_notification('Yatcobot: Someone mentioned you, you may won something!',
                                                '{} new mentions : \n {}'.format(len(posts), links))

            self.last_mention = posts[0]

    def run(self):
        """Run the bot as a daemon. This is blocking command"""

        self.scheduler.enter(TwitterConfig.get().scheduler.clear_queue_interval, 1, self.clear_queue)
        self.scheduler.enter(TwitterConfig.get().scheduler.rate_limit_update_interval, 2, self.client.update_ratelimits)
        self.scheduler.enter(TwitterConfig.get().scheduler.blocked_users_update_interval, 3, self.update_blocked_users)
        self.scheduler.enter(TwitterConfig.get().scheduler.check_mentions_interval, 4, self.check_new_mentions)
        self.scheduler.enter(TwitterConfig.get().scheduler.search_interval, 5, self.scan_new_contests)
        self.scheduler.enter_random(TwitterConfig.get().scheduler.retweet_interval,
                                    TwitterConfig.get().scheduler.retweet_random_margin, 6, self.enter_contest)

        # Init the program
        self.scheduler.run()

    def _get_original_tweet(self, post):
        """
        Checks if a post is a retweet and returns original tweet
        :param post: Post to check if its retweeted
        :return: post: If itsnt retweet it returns the argument, otherwise returns original tweet
        """
        if 'retweeted_status' in post:
            logger.debug('Tweet {} is a retweet'.format(post['id']))
            return post['retweeted_status']
        return post

    def _get_quoted_tweet(self, post):
        """
        Checks if a post is a quote of the original tweet
        Also the quote maybe is quoting another quote. So we follow quotes until we find the original or
        if we follow Config.max_quote_depth times
        :param post: The post to check if its a quote
        :return: If it isnt a quote the argument, otherwise the original tweet
        """
        for i in range(TwitterConfig.get().search.max_quote_depth):
            # If it hasnt quote return the post
            if 'quoted_status' not in post:
                return post

            quote = post['quoted_status']
            diff = difflib.SequenceMatcher(None, post['full_text'], quote['full_text']).ratio()
            # If the texts are similar continue
            if diff >= TwitterConfig.get().search.min_quote_similarity:
                logger.debug('{} is a quote, following to next post. Depth from original post {}'.format(post['id'], i))
                quote = self.client.get_tweet(quote['id'])
                # If its a quote of quote, get next quote and continue
                post = quote
                continue
            # Else return the last post
            break

        return post

    def _insert_post_to_queue(self, post):
        """
        Check if a post is wanted and add's it in the post queue
        :param post: The post to insert
        """
        # Get original tweet if retweeted
        post = self._get_original_tweet(post)

        # Get original post, if it is quoted
        post = self._get_quoted_tweet(post)

        # Filter retweeted
        if post['retweeted']:
            return

        # Filter ids in ignore list
        if post['id'] in self.ignore_list:
            return

        # Filter blocked users
        if post['user']['id'] in self.ignore_list:
            return

        # Filter posts with deleted quote
        # We check if there is a key 'is_a_quote_status' that is true but there isn't a quoted_status
        if 'is_quote_status' in post and post['is_quote_status'] and not 'quoted_status' in post:
            return

        # Insert if it doenst already exists
        if post['id'] not in self.post_queue:
            self.post_queue[post['id']] = post
            text = post['full_text'].replace('\n', '')
            text = (text[:75] + '..') if len(text) > 75 else text
            logger.debug("Added tweet to queue: id:{0} username:{1} text:{2}".format(post['id'],
                                                                                     post['user']['screen_name'],
                                                                                     text))

    def create_tweet_link(self, post):
        return "http://twitter.com/{}/status/{}".format(post['user']['screen_name'], post['id'])



import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def create_logger(level, file=None):
    # Create log outputs
    ch = logging.StreamHandler()

    # Log format
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s', "%Y-%m-%d %H:%M:%S")

    # Set logging format
    ch.setFormatter(formatter)

    # Set level per output

    ch.setLevel(level)

    if file is not None:
        fh = logging.FileHandler(file)
        fh.setLevel(level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    logger.addHandler(ch)



def create_keyword_mutations(keyword):
    """
    Given a keyword, create various mutations to be searched inside a post
    :param keyword: the base keyword of the mutations
    :return: list of mutation
    """
    mutations = list()
    keyword = keyword.strip()
    mutations.append(' {} '.format(keyword))
    mutations.append('{} '.format(keyword))
    mutations.append(' {}'.format(keyword))
    mutations.append('#{}'.format(keyword))
    mutations.append(',{}'.format(keyword))
    mutations.append('{},'.format(keyword))
    mutations.append('.{}'.format(keyword))
    mutations.append('{}.'.format(keyword))
    return mutations



import logging
from abc import ABC, abstractmethod

from .config import TwitterConfig
from .utils import create_keyword_mutations

logger = logging.getLogger(__name__)


class ActionABC(ABC):

    def __init__(self, client):
        self.client = client

    @abstractmethod
    def process(self, post):
        raise NotImplementedError("Action must implement process")


class Follow(ActionABC):
    """
    Checks if a contest needs follow to enter and follows the user

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if TwitterConfig.get().actions.follow.multiple:
            logger.warning('Experimental feature actions.follow.multiple is enabled.')
            logger.warning('Will follow every mentioned user in a post.')
            logger.warning('If unwanted behavior is observed, please open an issue on github along with the post id!')

    def process(self, post):
        text = post['full_text']
        keywords = sum((create_keyword_mutations(x) for x in TwitterConfig.get().actions.follow.keywords), [])
        if any(x in text.lower() for x in keywords):
            self.remove_oldest_follow()
            self.follow(post)

    def remove_oldest_follow(self):
        """
        If the follow limit is reached, unfollow the oldest follow
        """

        follows = self.client.get_friends_ids()

        if len(follows) > TwitterConfig.get().actions.follow.max_following:
            r = self.client.unfollow(follows[-1])
            logger.info('Unfollowed: {0}'.format(r['screen_name']))

    def follow(self, post):
        users_followed = list()

        # If multiple users is enabled follow all users mentioned
        if TwitterConfig.get().actions.follow.multiple:
            for user in post['entities']['user_mentions']:
                self.client.follow(user['screen_name'])
                users_followed.append(user['screen_name'])
                logger.info("Follow: {0}".format(user['screen_name']))

        # If op not already followed, follow
        if post['user']['screen_name'] not in users_followed:
            self.client.follow(post['user']['screen_name'])
            logger.info("Follow: {0}".format(post['user']['screen_name']))


class Favorite(ActionABC):
    """
    Checks if a contest needs favorite to enter, and favorites the post
    """

    def process(self, post):
        text = post['full_text']
        keywords = sum((create_keyword_mutations(x) for x in TwitterConfig.get().actions.favorite.keywords), [])
        if any(x in text.lower() for x in keywords):
            r = self.client.favorite(post['id'])
            logger.info("Favorite: {0}".format(post['id']))



import datetime
import logging
import re
import time

from TwitterAPI import TwitterAPI, constants

from .config import TwitterConfig

logger = logging.getLogger(__name__)


class TwitterClientException(Exception):
    pass


class TwitterClientRetweetedException(Exception):
    pass


class RateLimiterExpired(Exception):
    pass


class RateLimiter(dict):

    def refresh_limits(self, new_limits):
        '''
        Refreshes the limits with the new limits info that was getted from the twitter api
        :param new_limits: a dictionary with the new limits
        '''

        # delete old values
        self.clear()

        # flatten dictionay
        for value in new_limits.values():
            for key, limits in value.items():
                # replace parameters in url like :id :slug to :PARAM
                key = re.sub(r'\/:[a-z_]+', '/:PARAM', key)
                self[key] = limits

        # create percent
        for key in self.keys():
            self._calculate_percent_remaining(key)

        logger.debug('Ratelimits was refreshed')

    def check_limit(self, endpoint):
        '''
        Check if we exceded the api calls for a particular endpoint. If we did it pauses until the nex
        reset
        :param endpoint: the endpoint (api url) to be checked
        '''

        # Only GET methods have ratelimits
        if constants.ENDPOINTS[endpoint][0] == 'POST':
            return

        logger.debug('Checking limits for {}'.format(endpoint))

        endpoint_limits = self[self._get_internal_endpoint_name(endpoint)]

        # if over threshold sleep untill reset
        if endpoint_limits['percent'] < TwitterConfig.get().min_ratelimit_percent:

            reset_time = endpoint_limits['reset']
            now = int(datetime.datetime.now().strftime('%s'))
            wait = reset_time - now

            if wait > 0:
                logger.warning('Rate limit {}. Waiting for {} seconds'.format(endpoint_limits['percent'], wait))
                time.sleep(wait)
                raise RateLimiterExpired()

    def decrease_remaining(self, endpoint):
        '''
        Decreases the remaining calls of an endpoint
        :param endpoint: the endpoint to decrease
        '''
        endpoint = self._get_internal_endpoint_name(endpoint)
        if endpoint in self:
            self[endpoint]['remaining'] -= 1
            logger.debug('Decreased remainin usages of {}. New value {}'.format(endpoint, self[endpoint]['remaining']))
            self._calculate_percent_remaining(endpoint)

    def _calculate_percent_remaining(self, endpoint):
        '''
        Calculate the percent of remaining calls for an endpoint
        :param endpoint: The target endpoint to calculate
        '''
        self[endpoint]['percent'] = self[endpoint]['remaining'] / self[endpoint]['limit'] * 100

    def _get_internal_endpoint_name(self, endpoint):
        '''
        Maps an TwitterApi endpoint to the internal endpoint. This is usefull because
        TwitterApi has the endpoints without the leading '/' but the twitter returns
        a dictionary starting with '/'
        :param endpoint: the endpoint to convert
        :return: the internal endpoint
        '''
        return '/' + endpoint


class TwitterClient:

    def __init__(self, consumer_key, consumer_secret, access_token_key, access_token_secret):
        self.api = TwitterAPI(consumer_key, consumer_secret, access_token_key, access_token_secret)
        self.ratelimiter = RateLimiter()
        self.update_ratelimits(check_ratelimit=False)

    def search_tweets(self, query, limit, result_type='mixed', language=None):
        parameters = {'q': query, 'result_type': result_type, 'count': limit}
        if language is not None:
            parameters['lang'] = language
        r = self._api_call('search/tweets', parameters)
        return r['statuses']

    def get_tweet(self, post_id):
        return self._api_call('statuses/show/:{}'.format(post_id))

    def retweet(self, post_id):
        return self._api_call('statuses/retweet/:{}'.format(post_id))

    def get_friends_ids(self):
        return self._api_call('friends/ids')['ids']

    def follow(self, username):
        return self._api_call('friendships/create', {'screen_name': username})

    def unfollow(self, user_id):
        return self._api_call('friendships/destroy', {'user_id': user_id})

    def favorite(self, post_id):
        return self._api_call('favorites/create', {'id': post_id})

    def get_blocks(self):
        return self._api_call('blocks/ids')['ids']

    def get_mentions_timeline(self, count=None, since_id=None):
        parameters = dict()
        if count is not None:
            parameters['count'] = count

        if since_id is not None:
            parameters['since_id'] = since_id

        return self._api_call('statuses/mentions_timeline', parameters)

    def update_ratelimits(self, check_ratelimit=True):

        # ratelimit_check controls if before the api_call we check the ratelimit. Usefull to be false the first update
        r = self._api_call('application/rate_limit_status', check_ratelimit=check_ratelimit)['resources']

        self.ratelimiter.refresh_limits(r)

        # log ratelimit status
        logger.debug("Ratelimit status: {}".format({key: item['percent'] for key, item in self.ratelimiter.items()
                                                    if item['percent'] < 100}))

    def _api_call(self, request, parameters=None, check_ratelimit=True):

        # !Fixme: TwitterApi doenst have all the endpoints. Some will raise exception
        _, endpoint = self.api._get_endpoint(request)

        if check_ratelimit:

            try:
                self.ratelimiter.check_limit(endpoint)
            except RateLimiterExpired:
                self.update_ratelimits(check_ratelimit=False)

        # Add support for extended tweets
        if parameters is not None:
            parameters['tweet_mode'] = 'extended'
        else:
            parameters = {'tweet_mode': 'extended'}

        response = self.api.request(request, parameters).json()

        try:
            self._check_for_errors(response)
        except Exception as e:
            logger.error('Error during twitter api call {} (parameters: {})'.format(request, parameters))
            raise

        self.ratelimiter.decrease_remaining(endpoint)

        return response

    def _check_for_errors(self, response):
        # check for errors
        if 'errors' in response:
            for error in response['errors']:
                message = error['message']
                code = error['code']
                if message == 'You have already retweeted this Tweet.':
                    raise TwitterClientRetweetedException()
                logger.error('Twitter api error code:{} error:{}'.format(code, message))
            raise TwitterClientException()



import argparse
import logging
import os

from yatcobot import create_logger
from yatcobot.bot import Yatcobot
from yatcobot.config import TwitterConfig

logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(description='Yatcobot: a bot for entering twitter contests')
    parser.add_argument('--config', '-c', dest='config', default='config.yaml', help='Path of the config file')
    parser.add_argument('--ignore_list', '-i', dest='ignore_list', default='ignorelist', help='Path of the ignore file')
    parser.add_argument('--log', dest='logfile', default=None, help='Path of log file')
    parser.add_argument('--debug', dest='debug', action='store_true', help='Enable debug')

    args = parser.parse_args()

    # Create logger
    if args.debug:
        create_logger(logging.DEBUG, args.logfile)
    else:
        create_logger(logging.INFO, args.logfile)

    # Check for old config
    if args.config.endswith('.json') or (os.path.isfile('config.json') and not os.path.isfile('config.yaml')):
        logger.error("Config file format changed, please update your config to the new yaml format!")
        logger.error("Visit documentation for more info: https://yatcobot.readthedocs.io/en/master/config.html")
        exit(1)

    logger.info("Loading configuration")
    TwitterConfig.load(args.config)
    logger.info("Configuration loaded")
    logger.info("Starting")
    print_logo()
    bot = Yatcobot(args.ignore_list)
    bot.run()


def print_logo():
    # Figlet font: Standard
    logger.info(
        '''                                                        
                                  .:clcllcll;.                                  
                                ;xc.        .ld'                                
                              .k:     .'..     xd                               
                              x;    cdcc:cx,    xc                              
                             .WolllxO 0oko.XolllxO                              
                              X,...;0.llo,,O....lO                              
                             l0x    .lllclc     0O;                             
                            OOkOk'            :OOdXl                            
                         .dkK;K' col,.    .;ld, oO:Xdo                          
                         ,o.kWl    :XdX0XOdK.   .xNc.k.                         
                           ,l:    :kO:kOXd:0o,    cl'                           
                                 :O    kl   .K.                                 
                                 .llllllllllll.                                 
                                                                                
                                 _            _           _
                     _   _  __ _| |_ ___ ___ | |__   ___ | |_
                    | | | |/ _` | __/ __/ _ \| '_ \ / _ \| __|
                    | |_| | (_| | || (_| (_) | |_) | (_) | |_
                     \__, |\__,_|\__\___\___/|_.__/ \___/ \__|
                     |___/
 '''
    )



from abc import ABC
from collections import namedtuple, OrderedDict
from datetime import datetime
from statistics import mean, stdev

from dateutil import tz

from .config import TwitterConfig

Score = namedtuple('Score', ('id', 'score'))


class PostQueue(OrderedDict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.rating_methods = list()
        self.filter_methods = list()

        # Add rating methods
        if TwitterConfig.get().search.sort.by_keywords.enabled:
            self.rating_methods.append(RateByKeywords())

        if TwitterConfig.get().search.sort.by_age.enabled:
            self.rating_methods.append(RateByAge())

        if TwitterConfig.get().search.sort.by_retweets_count.enabled:
            self.rating_methods.append(RateByRetweetsCount())

        # Add filter methods

        if TwitterConfig.get().search.filter.min_retweets.enabled:
            self.filter_methods.append(FilterMinRetweets())

    def filter(self):
        """
         Will filter the queue based on the options provided in config
        """
        for method in self.filter_methods:
            method.filter(self)

    def sort(self):
        """
        Will sort the queue based on the options provided in config

        """
        combined_rates = self.combine_rates(*(method.get_rates(self) for method in self.rating_methods))

        sorted_rates = sorted((x for x in combined_rates.items()), key=lambda x: x[1], reverse=True)

        # add a score value to every post
        for post in sorted_rates:
            self[post[0]]['score'] = post[1]

        updated_queue = OrderedDict((x[0], self[x[0]]) for x in sorted_rates)
        self.clear()
        self.update(updated_queue)

    def combine_rates(self, *rates):
        """
        Combines the rates of multiple features to one final rate
        :param rates: a tuple of list of scores ( [Ascore1, Ascore2], [Bscore1,Bscore2])
        :return:Returns an a dict with final scores {id:Score}
        """
        combined_rates = {}
        for rates_list in rates:
            for rate in rates_list:
                current_score = combined_rates.get(rate.id, 0)
                current_score += rate.score
                combined_rates[rate.id] = current_score

        return combined_rates


# ----------------------------------------------------------------------------------------------------------------------

class RateABC(ABC):
    """
    Superclass of rating methods that are used for sorting the queue
    """

    config_name = ''

    def get_rates(self, queue):
        raise NotImplementedError('get_scores must be implemented')

    def normalize_scores(self, scores):
        """
        Computes the standardized score for a collection of scores
        :param scores: A list of scores [Score]
        :return:Standardized score
        """
        m = mean(x.score for x in scores)
        s = stdev(x.score for x in scores)

        normalized_scores = []

        # if standard deviation is 0 return 0 scores
        if s == 0:
            return [Score(x.id, 0) for x in scores]

        for x in scores:
            normalized_scores.append(Score(x.id, (x.score - m) / s))

        return normalized_scores


class RateByKeywords(RateABC):
    config_name = 'by_keywords'

    def get_rates(self, queue):
        """
        Gets a queue and returns the scores based on the keywords in text
        :param queue: The queue to score
        :return:A list of scores [Score]
        """
        rates = []

        # Find how many times each keyword appears in the text
        for post in queue.values():
            rate = 0
            text = post['full_text'].lower()

            for keyword in TwitterConfig.get().search.sort.by_keywords.keywords:
                keyword = keyword.lower()
                rate += text.count(keyword)

            rates.append(Score(post['id'], rate))

        norm_scores = self.normalize_scores(rates)

        return norm_scores


class RateByRetweetsCount(RateABC):
    config_name = 'by_retweets_count'

    def get_rates(self, queue):
        """
        Gets a queue and returns the scores based on the retweets of the post
        :param queue: The queue to score
        :return:A list of scores [Score]
        """
        rates = [Score(post_id, post['retweet_count']) for post_id, post in queue.items()]

        norm_rates = self.normalize_scores(rates)

        return norm_rates


class RateByAge(RateABC):
    config_name = 'by_age'

    def get_rates(self, queue):
        rates = []

        now = datetime.now(tz.tzutc())

        for post in queue.values():
            post_date = datetime.strptime(post['created_at'], '%a %b %d %H:%S:%M %z %Y')
            seconds = (post_date - now).total_seconds()
            rates.append(Score(post['id'], seconds))

        return self.normalize_scores(rates)


# ----------------------------------------------------------------------------------------------------------------------


class FilterABC(ABC):
    """
    Superclass of filtering methods for filtering the queue
    """

    def filter(self, queue):
        raise NotImplementedError('Subclasses must implement filter')


class FilterMinRetweets(FilterABC):
    def filter(self, queue):
        """
        Removes all post that have less rewteets than min_retweets defined at config
        :param queue: Posts Queue
        """
        ids_to_remove = list()

        for post_id, post in queue.items():
            if post['retweet_count'] < TwitterConfig.get().search.filter.min_retweets.number:
                ids_to_remove.append(post_id)

        for post_id in ids_to_remove:
            del queue[post_id]



import logging
import random
import sched
import time
from collections import namedtuple

logger = logging.getLogger(__name__)

# Task types
NormalTask = namedtuple('NormalTask', ['delay', 'priority', 'action'])
RandomTask = namedtuple('RandomTask', ['delay', 'delay_margin', 'priority', 'action'])


class PeriodicScheduler(sched.scheduler):
    """Schedules tasks to be called periodically"""

    def __init__(self, timefunc=time.time, delayfunc=time.sleep):
        # List of tasks that will be periodically be called
        self.tasks = []
        super().__init__(timefunc, delayfunc)

    def enter(self, delay, priority, action):
        """Inserts a new task in the scheduler. Tasks will be called regularly with period delay"""

        task = NormalTask(delay, priority, action)
        self.tasks.append(task)

    def enter_random(self, delay, delay_margin, priority, action):
        """
        Inserts a new random task in the scheduler. Tasks will be paused for random time
        from delay-delay_margin to delay+delay_margin
        """
        task = RandomTask(delay, delay_margin, priority, action)
        self.tasks.append(task)

    def run(self, blocking=True):
        for i in range(len(self.tasks)):
            self.run_task(i)

        super().run(blocking)

    def enter_task(self, index):
        """
        Inserts a task in the sched queue. We dont enter the task itself, rather the run_task with the index
        of taskas a parameter. We do it that way so we can reschedule it later
        :param index: the index of the task
        """
        task = self.tasks[index]
        if isinstance(task, NormalTask):
            super().enter(task.delay, task.priority, self.run_task, argument=(index,))
            logger.debug("Rescheduling {} for after {} seconds".format(task.action.__name__, task.delay))
        elif isinstance(task, RandomTask):
            delay = random.randint(task.delay - task.delay_margin, task.delay + task.delay_margin)
            logger.debug("Rescheduling {} for after {} seconds".format(task.action.__name__, delay))
            super().enter(delay, task.priority, self.run_task, argument=(index,))

    def run_task(self, index):
        """
        Runs a task and reschedules it again
        :param index:  the index of the task to run
        """
        self.enter_task(index)
        try:
            logger.debug("Scheduler is calling: {}".format(self.tasks[index].action.__name__))
            self.tasks[index].action()
        except Exception as e:
            logger.error("Exception in scheduled task :{}".format(self.tasks[index].action.__name__), exc_info=True)
            pass



import logging
from abc import ABCMeta, abstractmethod, abstractstaticmethod, abstractclassmethod

from pushbullet import PushBullet

from .config import TwitterConfig, NotifiersConfig

logger = logging.getLogger(__name__)


class NotificationService:

    def __init__(self):
        self.active_notifiers = []
        self._intialize_notifiers()

    def send_notification(self, title, message):
        """Sends a message to all enabled notifiers"""
        for notifier in self.active_notifiers:
            notifier.notify(title, message)

    def _intialize_notifiers(self):
        """Adds all enabled notifiers in the active_notifiers list"""
        for cls in AbstractNotifier.__subclasses__():
            if cls.is_enabled():
                self.active_notifiers.append(cls.from_config())

    def is_enabled(self):
        """Checks if any notifier is enabled"""

        if len(self.active_notifiers) > 0:
            return True
        return False


class AbstractNotifier(metaclass=ABCMeta):
    """
    Abstract class that all methods that want to notify the user must derive from
    """

    @abstractmethod
    def notify(self, title, message):
        """Sends the message to the user"""

    @staticmethod
    @abstractstaticmethod
    def is_enabled():
        """
        Returns if its enabled or not
        :return: Bool
        """

    @classmethod
    @abstractclassmethod
    def from_config(cls):
        """
        Instatiates a Notifier class using parameters from config
        :return:
        """


class PushbulletNotifier(AbstractNotifier):

    def __init__(self, api_key):
        self.pb = PushBullet(api_key)

    def notify(self, title, message):
        self.pb.push_note(title, message)

    @staticmethod
    def is_enabled():
        if NotifiersConfig.get().pushbullet.enabled:
            return True
        return False

    @classmethod
    def from_config(cls):
        return cls(NotifiersConfig.get().pushbullet.token)



class IgnoreList(set):
    """
    A list like object that loads contents from a file and everything that is appended here gets also
    appended in the file
    """

    def __init__(self, filename):
        self.filename = filename
        self.load_file()

    def append(self, p_object):
        """
        Adds an id to the ignore list and apeends it to the file
        :param p_object: int the id
        """
        if p_object not in self:
            self.append_file(p_object)
        super().add(p_object)

    def load_file(self):
        """
        loads an ignore list from a file
        """
        with open(self.filename, 'a+') as f:
            f.seek(0)
            self.update(int(x) for x in f.read().splitlines())

    def append_file(self, p_object):
        """
        append an id to the file
        :param p_object: int the id
        """
        with open(self.filename, 'a+') as f:
            f.write(str(p_object) + '\n')



import os.path
import logging

import confuse
import pkg_resources
import yaml


logger = logging.getLogger(__name__)


class Config:
    template = {
        'twitter': {
            'consumer_key': confuse.String(),
            'consumer_secret': confuse.String(),
            'access_token_key': confuse.String(),
            'access_token_secret': confuse.String(),
            'min_ratelimit_percent': confuse.Integer(),

            'search': {
                'queries': confuse.TypeTemplate(list),
                'max_queue': confuse.Integer(),
                'max_quote_depth': confuse.Integer(),
                'min_quote_similarity': confuse.Number(),
                'filter': {
                    'min_retweets': {
                        'enabled': confuse.TypeTemplate(bool),
                        'number': confuse.Integer()
                    }
                },
                'sort': {
                    'by_keywords': {
                        'enabled': confuse.TypeTemplate(bool),
                        'keywords': confuse.StrSeq()
                    },
                    'by_age': {
                        'enabled': confuse.TypeTemplate(bool),
                    },
                    'by_retweets_count': {
                        'enabled': confuse.TypeTemplate(bool),
                    }
                }
            },

            'actions': {
                'follow': {
                    'enabled': confuse.TypeTemplate(bool),
                    'keywords': confuse.StrSeq(),
                    'max_following': confuse.Integer(),
                    'multiple': confuse.TypeTemplate(bool)
                },
                'favorite': {
                    'enabled': confuse.TypeTemplate(bool),
                    'keywords': confuse.StrSeq()
                },
            },

            'scheduler': {
                'search_interval': confuse.Integer(),
                'retweet_interval': confuse.Integer(),
                'retweet_random_margin': confuse.Integer(),
                'blocked_users_update_interval': confuse.Integer(),
                'clear_queue_interval': confuse.Integer(),
                'rate_limit_update_interval': confuse.Integer(),
                'check_mentions_interval': confuse.Integer(),
            },
        },
        'notifiers': {
            'pushbullet': {
                'enabled': confuse.TypeTemplate(bool),
                'token': confuse.String()
            }
        }
    }

    _valid = None

    @staticmethod
    def get():
        """
        Gets the static config object
        :return:
        """
        if Config._valid is None:
            raise ValueError("Configuration not loaded")
        return Config._valid

    @staticmethod
    def load(filename=None):
        """
        Loads a file and imports the settings
        :param filename: the file to import
        """
        config = confuse.LazyConfig('Yatcobot', __name__)

        # Add default config when in egg (using this way because egg is breaking the default way)
        if len(config.sources) == 0:
            default_config_text = pkg_resources.resource_string("yatcobot", "config_default.yaml")
            default_config = confuse.ConfigSource(yaml.load(default_config_text, Loader=confuse.Loader),
                                                  'pkg/config_default.yaml',
                                                  True)
            config.add(default_config)

        # Add user specified config
        if filename is not None and os.path.isfile(filename):
            config.set_file(filename)

        logger.info('Loading config files (From highest priority to lowest):')
        for i, config_source in enumerate(config.sources):
            logger.info('{}: Path: {}'.format(i, config_source.filename))
        Config._valid = config.get(Config.template)


class TwitterConfig(Config):

    @staticmethod
    def get():
        return super(TwitterConfig, TwitterConfig).get().twitter


class NotifiersConfig(Config):

    @staticmethod
    def get():
        return super(NotifiersConfig, NotifiersConfig).get().notifiers



#!/usr/bin/env python
from os import path
from setuptools import setup, find_packages
from pip.req import parse_requirements


requirements = parse_requirements('requirements.txt', session='hack')
requirements = [str(ir.req) for ir in requirements]
root_path = path.abspath(path.dirname(__file__))
with open(path.join(root_path, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='Yatcobot',

    version='2.1.2',

    description='The best bot for searching twitter contests and automatically retweet them',

    long_description=long_description,  # Optional

    url='https://github.com/buluba89/Yatcobot',

    author='John Buluba',

    classifiers=[
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='twitter bot contest retweet',

    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'utils']),  # Required

    package_data={  # Optional
        'yatcobot': ['config_default.yaml'],
    },

    install_requires=requirements,
    setup_requires=['pipenv'],

    entry_points={  # Optional
        'console_scripts': [
            'yatcobot=yatcobot.cli:main',
        ],
    },

    project_urls={
        'Source': 'https://github.com/buluba89/Yatcobot',
    },
)
